from typing import TYPE_CHECKING, List, Optional

import numpy as np
import pandas as pd
from sklearn.metrics import roc_auc_score

if TYPE_CHECKING:
    from .calculator import Calculator


def calculate_score(x: pd.DataFrame, target_column: str, k: str) -> float:
    """
    Helper function to calculate roc_auc_score and handle exceptions.

    :param x: Input DataFrame.
    :param target_column: Column name to use for target values in AUC calculation.
    :param k: Key used for the boundary dictionary and to format column names.
    :return: Calculated score or 0.5 in case of exceptions.
    """
    try:
        return float(roc_auc_score(x[f"{target_column}_lt_{k}"], x["overall_score"]))
    except:
        return 0.5


def calculate_woauc(
    calculator: "Calculator",
    groupby: Optional[str],
    target_column: str,
    weights_for_equation: List[float],
    weights_for_groups: Optional[pd.Series] = None,
) -> List[float]:
    """
    Calculate weighted ordinal user AUC.

    :param calculator: Calculator object that contains the data and methods for calculation.
    :param groupby: Column name to group data by, or None for no grouping.
    :param target_column: Column name to use for target values in AUC calculation.
    :param weights_for_equation: List of weights for equation.
    :param weights_for_groups: Series of weights for groups, or None.
    :return: List of calculated weighted ordinal user AUC values.
    """
    woauc_indices = calculator.woauc_dict[target_column]
    woauc = []
    sampler = calculator.samplers[target_column]

    for k, _ in sampler.boundary_dict.items():
        if groupby is not None:
            grouped = (
                calculator.df.loc[woauc_indices]
                .groupby(groupby)
                .apply(lambda x: calculate_score(x, target_column, k))
            )

            if weights_for_groups is not None:
                counts_sorted = weights_for_groups.loc[grouped.index]
                paritial_auc = float(np.average(grouped, weights=counts_sorted.values))
            else:
                paritial_auc = float(np.mean(grouped))
        else:
            paritial_auc = float(
                roc_auc_score(
                    calculator.df.loc[woauc_indices][f"{target_column}_lt_{k}"],
                    calculator.df.loc[woauc_indices]["overall_score"],
                )
            )
        woauc.append(paritial_auc)
    return woauc



from typing import List, Optional, Tuple

import matplotlib.patches as mpatches
import numpy as np
from matplotlib import pyplot as plt


class TrisurfPloter:
    """TrisurfPloter class for plotting trisurf."""

    def __init__(
        self,
        surf_data_group: List[Tuple[np.ndarray, np.ndarray, np.ndarray]],
        surf_labels: List[str],
        xyz_labels: List[str],
        cmap_colors: List[str] = ["Blues", "Greens", "Oranges"],
        alphas: List[float] = [0.5, 0.6, 0.7],
        legend_loc: str = "lower center",
        view_init: Tuple[float, float] = (10, 35),
        save_fig: bool = False,
        file_tag: Optional[float] = None,
        file_type: str = "pdf",
    ) -> None:
        """Initialize TrisurfPloter.

        :param surf_data_group: surf data group
        :param surf_labels: surf labels
        :param cmap_colors: cmap colors
        :param alphas: alphas
        :param legend_loc: legend location
        :param view_init: view init
        :param save_fig: whether to save the figure
        :param file_tag: file tag
        :param file_type: file type
        """
        self.surf_data_group = surf_data_group
        self.surf_labels = surf_labels
        self.xyz_labels = xyz_labels
        self.cmap_colors = cmap_colors
        self.alphas = alphas
        self.legend_loc = legend_loc
        self.view_init = view_init
        self.save_fig = save_fig
        self.file_tag = file_tag
        self.file_type = file_type
        self.surf_num = len(surf_data_group)
        self.plot_trisurfs()

    def init_figure(self, grid_color: str = "white") -> None:
        """Initialize figure."""
        self.fig = plt.figure()
        self.ax = self.fig.add_subplot(111, projection="3d")
        self.ax.grid(color=grid_color)

    def get_color_patches(self) -> List[mpatches.Patch]:
        """Get color patches for legend.

        :return: color patches
        """
        handles = []
        for surf, surf_label in zip(self.surfs, self.surf_labels):
            handles.append(
                mpatches.Patch(color=surf.get_facecolor()[-1], label=surf_label)
            )
        return handles

    def set_xyz_labels(self) -> None:
        """Set xyz labels.""" ""
        self.ax.set_xlabel(self.xyz_labels[0])
        self.ax.set_ylabel(self.xyz_labels[1])
        self.ax.set_zlabel(self.xyz_labels[2])

    def plot_trisurfs(self) -> None:
        """Plot trisurfs."""
        self.init_figure()
        self.surfs = []
        for surf_data, cmap_color, alpha in zip(
            self.surf_data_group, self.cmap_colors, self.alphas
        ):
            self.surfs.append(
                self.ax.plot_trisurf(
                    surf_data[0].ravel(),
                    surf_data[1].ravel(),
                    surf_data[2].ravel(),
                    cmap=cmap_color,
                    edgecolor="none",
                    antialiased=True,
                    alpha=alpha,
                )
            )
        handles = self.get_color_patches()
        self.set_xyz_labels()
        self.ax.legend(
            handles=handles, loc=self.legend_loc, frameon=False, ncol=self.surf_num
        )
        elev, azim = self.view_init
        self.ax.view_init(elev, azim)
        if self.save_fig:
            self.fig.savefig(f"trisurf_{self.file_tag}.{self.file_type}")
        plt.show()



from .base_calculator import BaseCalculator
from .calculator import Calculator
from .logarithm_pca_calculator import LogarithmPCACalculator

__all__ = [
    "BaseCalculator",
    "Calculator",
    "LogarithmPCACalculator",
]



from typing import List, Optional

import numpy as np
import pandas as pd

from .base_calculator import BaseCalculator


class Calculator(BaseCalculator):
    """Calculator class for calculating various metrics."""

    def __init__(
        self,
        df: pd.DataFrame,
        selected_columns: List[str],
        equation_type: str = "product",
        weights_for_groups: Optional[pd.Series] = None,
        equation_eval_str: Optional[str] = None,
    ) -> None:
        """Initialize Calculator.

        :param df: dataframe
        :param selected_columns: selected columns
        :param weights_for_groups: weights for group
        :param equation_eval_str: equation eval string
        """
        self.df = df
        self.df_len = len(self.df)
        self.equation_eval_str = equation_eval_str
        self.equation_type = equation_type
        self.samplers: dict = {}
        self.selected_columns = selected_columns
        self.selected_values = self.df[selected_columns].values
        self.woauc_dict: dict = {}
        self.bin_mappings: dict = {}
        if weights_for_groups is None:
            self.weights_for_groups = pd.Series(
                np.ones(len(self.df)), index=self.df.index
            )
        else:
            self.weights_for_groups = weights_for_groups
        self.value_scale()

    def value_scale(self) -> None:
        """
        Calculates the negative average log10 magnitude of absolute values for selected columns in the dataframe,
        storing the result in `self.value_scales`.
        """
        dataframe = self.df[self.selected_columns].abs()
        magnitudes = np.log10(dataframe.values)

        avg_magnitude = np.nanmean(magnitudes, axis=0)
        magnitudes = [-magnitude for magnitude in avg_magnitude]

        self.value_scales = np.asarray(magnitudes)

    def get_overall_score(
        self,
        weights_for_equation: List[float],
    ) -> None:
        """Calculate overall score.

        :param powers_for_equation: powers for equation
        :param first_order_weights: first order weights
        """
        if len(weights_for_equation) == 2 * len(self.selected_columns):
            powers_for_equation = weights_for_equation[: len(self.selected_columns)]
            first_order_weights = weights_for_equation[len(self.selected_columns) :]
            self.df["overall_score"] = np.prod(
                (1 + np.asarray(first_order_weights) * np.asarray(self.selected_values))
                ** powers_for_equation,
                axis=1,
            )
        elif self.equation_type == "product":
            self.df["overall_score"] = np.prod(
                self.selected_values**weights_for_equation, axis=1
            )
        elif self.equation_type == "sum":
            weights_array = np.array(weights_for_equation).reshape(-1, 1)
            self.df["overall_score"] = self.selected_values @ weights_array

        elif self.equation_type == "free_style":
            columns = [
                self.selected_values[:, i] for i in range(self.selected_values.shape[1])
            ]
            local_dict = {"weights": weights_for_equation, "columns": columns}
            if self.equation_eval_str is not None:
                self.df["overall_score"] = eval(
                    self.equation_eval_str, globals(), local_dict
                )
            else:
                raise ValueError("equation_eval_str is not defined.")

    def create_score_columns(
        self, boundary_dict: dict, score_column: str = "score"
    ) -> None:
        """Create score columns.

        :param boundary_dict: boundary dict
        :param score_column: score column to be converted
        """
        for k, _ in boundary_dict.items():
            self.df[f"{score_column}_lt_{k}"] = (self.df[score_column] >= k).astype(int)

    def initialize_fq_sampler(
        self,
        sample_size: int,
        score_column: str,
        slice_from: Optional[float] = None,
        slice_to: Optional[float] = None,
        log_scale: Optional[bool] = True,
        laplace_smoothing: Optional[bool] = True,
    ) -> None:
        """Initialize frequency sampler."""
        from ..sampling.frequency_sampler import FrequencySampler

        sampler = FrequencySampler(
            sample_size=sample_size,
            data=self.df[score_column],
            slice_from=slice_from,
            slice_to=slice_to,
            log_scale=log_scale,
            laplace_smoothing=laplace_smoothing,
        )
        self.create_score_columns(
            boundary_dict=sampler.sample(), score_column=score_column
        )
        self.samplers[score_column] = sampler
        slice_from_condition = (
            pd.Series(True, index=self.df.index)
            if slice_from is None
            else (self.df[score_column] >= slice_from)
        )
        slice_to_condition = (
            pd.Series(True, index=self.df.index)
            if slice_to is None
            else (self.df[score_column] <= slice_to)
        )
        self.woauc_dict[score_column] = self.df[
            slice_from_condition & slice_to_condition
        ].index



import logging
from abc import ABCMeta, abstractmethod
from typing import Dict, Optional, Union

import numpy as np
import optuna
from pydantic import BaseModel

from ..evaluation import Calculator, LogarithmPCACalculator
from .set_path import ensure_study_directory


class BaseObjectiveConfig(BaseModel):
    direction: Optional[str] = None
    formula: Optional[str] = None
    first_order: Optional[bool] = False
    power: Optional[bool] = True
    dirichlet: Optional[bool] = True
    weights_num: Optional[int] = None
    study_name: Optional[str] = None
    study_path: Optional[str] = None
    save_study: Optional[bool] = True


class BaseObjective(metaclass=ABCMeta):
    """
    BaseObjective serves as an abstract base class for objective optimization.
    It provides core functionalities for setting up and optimizing objectives
    using the `optuna` library.

    Attributes:
        calculator (Calculator): An instance of Calculator for various evaluations.
        direction (Optional[str]): Direction for optimization (e.g., "minimize" or "maximize").
        formula (Optional[str]): Mathematical formula representing the objective.
        first_order (bool): Whether to use first order optimization or not. Default is False.
        power (bool): If True, includes power in the optimization. Default is True.
        dirichlet (bool): If True, uses Dirichlet distribution for optimization. Default is True.
        weights_num (Optional[int]): Number of weights for optimization.
        study_name (Optional[str]): Name of the study.
        study_path (Optional[str]): Path to the study directory.
        full_path (str): Full path combining study_path and study_name.
        study (Study): Optuna study object for optimization.
        logger (logging.Logger): Logger object for logging optimization progress.

    Methods:
        build_logger(process_id) -> None:
            Constructs a logger for the optimization process.
        objective(trial) -> float:
            Abstract method for objective function to be overridden in derived classes.
        optimize(n_trials) -> None:
            Optimize the objective for a specified number of trials.
    """

    def __init__(
        self,
        calculator: Union[Calculator, LogarithmPCACalculator],
        direction: Optional[str] = None,
        formula: Optional[str] = None,
        first_order: Optional[bool] = False,
        power: Optional[bool] = True,
        dirichlet: Optional[bool] = True,
        weights_num: Optional[int] = None,
        study_name: Optional[str] = None,
        study_path: Optional[str] = None,
        save_study: Optional[bool] = True,
        config: Optional[Dict] = None,
    ) -> None:
        """
        Initializes the BaseObjective class with necessary parameters.

        Args:
            calculator (Calculator): Calculator for evaluations.
            direction (str): Direction of optimization.
            weights_num (int): Number of weights.
            formula (str): Formula for objective.
            first_order (bool, optional): Use first order optimization or not. Defaults to False.
            power (bool, optional): Include power in optimization. Defaults to True.
            dirichlet (bool, optional): Use Dirichlet distribution. Defaults to True.
            study_name (Optional[str], optional): Name of the study. Defaults to None.
            study_path (Optional[str], optional): Path to the study directory. Defaults to None.
        """
        self.calculator = calculator
        if config is not None:
            self.config = BaseObjectiveConfig(**config)
        else:
            self.config = BaseObjectiveConfig(
                direction=direction,
                formula=formula,
                first_order=first_order,
                power=power,
                dirichlet=dirichlet,
                weights_num=weights_num,
                study_name=study_name,
                study_path=study_path,
                save_study=save_study,
            )

        self.direction = self.config.direction
        self.formula = self.config.formula
        self.first_order = self.config.first_order
        self.power = self.config.power
        self.dirichlet = self.config.dirichlet
        self.weights_num = self.config.weights_num
        self.study_name = self.config.study_name
        self.study_path = self.config.study_path
        self.save_study = self.config.save_study

        self.full_path = ensure_study_directory(self.study_path, self.study_name)
        storage = optuna.storages.RDBStorage(
            url=f"sqlite:///{self.full_path}/paradance_storage.db",
            engine_kwargs={"connect_args": {"timeout": 60}},
        )
        self.study = optuna.create_study(
            direction=self.direction,
            study_name=self.study_name,
            storage=storage,
            load_if_exists=True,
        )
        if self.weights_num is None:
            self.weights_num = self.get_weights_num()
        self.best_params: np.ndarray = np.zeros(self.weights_num)

    def build_logger(self, process_id: Optional[int] = None) -> None:
        """
        Constructs a logger for the optimization process.

        Args:
            process_id (Optional[int], optional): ID of the process, used for creating unique log filenames.
                                                  Defaults to None.
        """

        if process_id is not None:
            log_filename = f"{self.full_path}/paradance_{process_id}.log"
        else:
            log_filename = f"{self.full_path}/paradance.log"

        file_handler = logging.FileHandler(log_filename)
        file_handler.setLevel(logging.INFO)
        self.logger = optuna.logging.get_logger(
            f"paradance_{process_id}" if process_id else "optuna"
        )
        self.logger.addHandler(file_handler)

    @abstractmethod
    def objective(self, trial: optuna.trial.Trial) -> float:
        """
        Abstract method for the objective function. Must be overridden in derived classes.

        Args:
            trial (Trial): Optuna trial instance.

        Returns:
            float: Objective value for the given trial.
        """
        raise NotImplementedError

    def get_weights_num(self) -> int:
        """
        Returns the number of weights.

        Returns:
            int: Number of weights.
        """
        return len(self.calculator.selected_columns)

    def optimize(self, n_trials: int) -> None:
        """
        Optimizes the objective for a set number of trials.

        Args:
            n_trials (int): Number of trials for optimization.
        """
        self.build_logger()
        self.study.optimize(self.objective, n_trials=n_trials)



import os
from typing import List, Optional, Tuple

import imageio
import matplotlib.pyplot as plt
import pandas as pd
from matplotlib_venn import venn3

from .venn_base import BaseVennPloter


class Venn3Ploter(BaseVennPloter):
    """
    Class to plot a venn diagram with 4 groups
    """

    def __init__(
        self, dataframe: pd.DataFrame, column1_name: str, column2_name: str
    ) -> None:
        super().__init__(dataframe, column1_name, column2_name)

    def get_conditions(self) -> Tuple:
        """
        Returns the conditions for the venn diagram
        """
        both0 = (self.column1_data == 0) & (self.column2_data == 0)
        only_column1 = (self.column1_data == 1) & (self.column2_data == 0)
        only_column2 = (self.column1_data == 0) & (self.column2_data == 1)
        both1 = (self.column1_data == 1) & (self.column2_data == 1)
        return (both0, only_column1, only_column2, both1)

    def get_group_count_ratios(self) -> Tuple:
        """
        Returns the ratios of the groups based on the count
        """
        both0, only_column1, only_column2, both1 = self.get_conditions()
        both0_ratio = sum(both0) / self.df_len
        only_column1_ratio = sum(only_column1) / self.df_len
        only_column2_ratio = sum(only_column2) / self.df_len
        both1_ratio = sum(both1) / self.df_len
        return (both0_ratio, only_column1_ratio, only_column2_ratio, both1_ratio)

    def get_group_value_ratios(
        self,
        value_column_name: str,
        lower_bound: Optional[float] = None,
        upper_bound: Optional[float] = None,
    ) -> Tuple:
        """
        Returns the ratios of the groups based on the value

        value_column_name: name of the column with the values
        lower_bound: lower bound of the values
        upper_bound: upper bound of the values
        """
        if lower_bound is None:
            greater_than_lower_bound = True
        else:
            greater_than_lower_bound = self.df[value_column_name] > lower_bound
        if upper_bound is None:
            lower_than_upper_bound = True
        else:
            lower_than_upper_bound = self.df[value_column_name] < upper_bound

        both0, only_column1, only_column2, both1 = self.get_conditions()
        both0 = both0 & greater_than_lower_bound & lower_than_upper_bound
        only_column1 = only_column1 & greater_than_lower_bound & lower_than_upper_bound
        only_column2 = only_column2 & greater_than_lower_bound & lower_than_upper_bound
        both1 = both1 & greater_than_lower_bound & lower_than_upper_bound
        both0_sum = self.df[both0][value_column_name].sum()
        only_column1_sum = self.df[only_column1][value_column_name].sum()
        only_column2_sum = self.df[only_column2][value_column_name].sum()
        both1_sum = self.df[both1][value_column_name].sum()
        value_sum = both0_sum + only_column1_sum + only_column2_sum + both1_sum
        both0_sum_ratio = both0_sum / value_sum
        only_column1_sum_ratio = only_column1_sum / value_sum
        only_column2_sum_ratio = only_column2_sum / value_sum
        both1_sum_ratio = both1_sum / value_sum
        return (
            float(both0_sum_ratio),
            float(only_column1_sum_ratio),
            float(only_column2_sum_ratio),
            float(both1_sum_ratio),
        )

    def plot_ratio_venn(
        self,
        ratios: tuple,
        save_fig: bool = False,
        file_tag: Optional[float] = None,
        file_type: str = "pdf",
    ) -> None:
        """
        Plots the venn diagram with the ratios

        ratios: ratios of the groups
        save_fig: whether to save the figure
        file_tag: tag to add to the file name
        file_type: file type to save the figure
        """
        both0_ratio, only_column1_ratio, only_column2_ratio, both1_ratio = ratios
        v = venn3(
            subsets=(
                0,
                0,
                0,
                both0_ratio,
                only_column1_ratio,
                only_column2_ratio,
                both1_ratio,
            ),
            set_labels=(self.column1_name, self.column2_name, "all"),
        )
        v.get_label_by_id("101").set_text(
            f"{only_column1_ratio:.3%}\n{self.column1_name} && !{self.column2_name}"
        )
        v.get_label_by_id("011").set_text(
            f"{only_column2_ratio:.3%}\n!{self.column1_name} && {self.column2_name}"
        )
        v.get_label_by_id("001").set_text(
            f"\n\n\n\n\n\n{both0_ratio:.3%}\n!{self.column1_name} && !{self.column2_name}"
        )
        v.get_label_by_id("111").set_text(
            f"\n\n\n\n\n\n\n\n{both1_ratio:.3%}\n{self.column1_name} && {self.column2_name}"
        )

        v.get_patch_by_id("101").set_color("orange")
        v.get_patch_by_id("011").set_color("green")
        v.get_patch_by_id("111").set_color("yellowgreen")
        v.get_patch_by_id("001").set_color("skyblue")

        if save_fig:
            if file_tag is None:
                file_name = f"ratio_venn.{file_type}"
            else:
                text_str: str = f"upper bound: {file_tag}"
                plt.text(
                    0.90,
                    0.05,
                    text_str,
                    transform=plt.gca().transAxes,
                    fontsize=10,
                    verticalalignment="top",
                )
                file_name = f"ratio_venn_{file_tag}.{file_type}"
            plt.savefig(file_name, format=file_type)

    def plot_value_ratio_venn_with_upper_bounds(
        self,
        value_column_name: str,
        upper_bounds: List[float],
        duration: Optional[int] = 300,
    ) -> None:
        filenames: List[str] = []
        for upper_bound in upper_bounds:
            filename = f"ratio_venn_{upper_bound}.png"
            filenames.append(filename)

        for upper_bound in upper_bounds:
            self.plot_value_ratio_venn(
                value_column_name,
                lower_bound=None,
                upper_bound=upper_bound,
                save_fig=True,
                file_type="png",
            )
            plt.close()
        with imageio.get_writer(
            f"ratio_venn_bounds_from_{upper_bounds[0]}_to_{upper_bounds[-1]}.gif",
            mode="I",
            duration=duration,
            loop=0,
        ) as writer:
            for filename in filenames:
                image = imageio.v2.imread(filename, pilmode="RGBA")
                writer.append_data(image)
        for filename in set(filenames):
            os.remove(filename)



from typing import TYPE_CHECKING, List, Optional, Union

import numpy as np
import pandas as pd
from scipy.stats import kendalltau

if TYPE_CHECKING:
    from .calculator import Calculator

from typing import List

import numpy as np


def map_to_bins(
    data: Union[List[float], np.ndarray], num_bins: int = 100
) -> np.ndarray:
    """
    Map data to equal-frequency bins.

    :param data: Union of list of floats or np.ndarray. The input data to be binned.
    :param num_bins: Optional integer, default 100. The number of bins to use for the mapping.
    :returns: np.ndarray. The binned data.
    """
    data = np.array(data, dtype=float)

    if len(data) == 0:
        return np.array([])

    if np.all(data == 0):
        return np.zeros_like(data)

    non_zero_data: np.ndarray = data[data != 0]
    num_bins = min(num_bins, len(non_zero_data))
    quantiles = np.linspace(0, 1, num_bins + 1)[1:-1]
    non_zero_bins = np.digitize(non_zero_data, np.quantile(non_zero_data, quantiles))

    bins = np.zeros_like(data)
    bins[data != 0] = non_zero_bins + 1

    return bins


def calculate_tau(
    calculator: "Calculator",
    target_column: str,
    groupby: Optional[str],
    weights_for_groups: Optional[pd.Series] = None,
    num_bins: Optional[Union[int, float]] = 100,
) -> float:
    """
    Calculate the Kendall's Tau using binned data.

    :param calculator: Calculator object. Stores the DataFrame and any existing bin mappings.
    :param target_column: String. The column name in the DataFrame that you want to target.
    :param groupby: Optional string. The column name to group by.
    :param weights_for_groups: Optional pd.Series. Weights for each group.
    :param num_bins: Optional integer, default 100. The number of bins to use for the mapping.
    :returns: float. The calculated Kendall's Tau. Normalized to [0, 1].
    """
    if num_bins is None:
        num_bins = 100
    else:
        num_bins = int(num_bins)
    if target_column in calculator.bin_mappings:
        label_bins = calculator.bin_mappings[target_column]
    else:
        label_bins = map_to_bins(pd.Series(calculator.df[target_column]), num_bins)
        calculator.bin_mappings[target_column] = label_bins
    if groupby is not None:
        calculator.df[f"{target_column}_bin"] = label_bins
        grouped = calculator.df.groupby(groupby).apply(
            lambda x: float(
                kendalltau(x[f"{target_column}_bin"], x["overall_score"])[0]
            )
        )
        if weights_for_groups is not None:
            counts_sorted = weights_for_groups.loc[grouped.index]
            tau = float(np.average(grouped, weights=counts_sorted.values))
        else:
            tau = float(np.mean(grouped))
    else:
        overall_score_bins = map_to_bins(calculator.df["overall_score"], num_bins)
        tau, _ = kendalltau(label_bins, overall_score_bins)
        normalized_tau = tau * 0.5 + 0.5
    return normalized_tau



import os
from typing import List, Optional, Tuple, Union

import imageio
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from tqdm import tqdm


class LorenzCurveGini:
    """
    This class provides methods to compute Gini coefficient and plot Lorenz curve.
    """

    def __init__(self, data: Union[pd.Series, List[float]]):
        """
        Initialize with a list of data.

        :param data: a list of floats
        """
        self.data = sorted(data)
        self.unique_data = list(set(self.data))

    def slice_data(
        self, lower_bound: Optional[float], upper_bound: Optional[float]
    ) -> Union[pd.Series, List[float]]:
        """
        Slice data from lower_bound to upper_bound.

        :param lower_bound: lower bound of the data
        :param upper_bound: upper bound of the data
        :return: a list of floats
        """
        data = []
        if lower_bound is None and upper_bound is None:
            data = self.data
        elif lower_bound is None and upper_bound is not None:
            data = [x for x in self.data if x <= upper_bound]
        elif upper_bound is None and lower_bound is not None:
            data = [x for x in self.data if lower_bound <= x]
        elif lower_bound is not None and upper_bound is not None:
            data = [x for x in self.data if lower_bound <= x <= upper_bound]
        return data

    def get_bounds(
        self,
        num_quantiles: int,
        slice_from: Optional[float] = None,
        slice_to: Optional[float] = None,
        unique_bounds: bool = False,
    ) -> List[Union[int, float]]:
        """
        Get bounds from the data according to quantiles.

        :param num_quantiles: number of quantiles
        :param lower_bound: lower bound of the data
        :param upper_bound: upper bound of the data
        :param unique_bounds: whether to return unique bounds
        :return: a list of floats
        """

        data = self.slice_data(slice_from, slice_to)
        quantiles = np.linspace(0, 1, num_quantiles + 2)[1:-1]
        bounds = np.percentile(data, quantiles * 100).astype(int)

        if unique_bounds:
            bounds = np.unique(bounds)

        return list(bounds)

    @staticmethod
    def gini_coefficient(data: Union[pd.Series, List[float]]) -> float:
        """
        Compute Gini coefficient.

        :param data: a list of floats
        :return: Gini coefficient as a float.
        """
        n = len(data)
        index = np.arange(1, n + 1)
        gini = (2 * np.sum(index * data) - (n + 1) * np.sum(data)) / (n * np.sum(data))
        return float(gini)

    @staticmethod
    def plot_lorenz_curve(
        data: Union[pd.Series, List[float]],
        save_fig: bool = False,
        file_tag: Optional[float] = None,
        file_type: str = "pdf",
    ) -> None:
        """
        Plot Lorenz curve.

        :param data: a list of floats
        :param save_fig: whether to save the figure
        :param file_tag: file tag to save the figure, which is the lower bound of the data
        :param file_type: file type to save the figure
        :return: None
        """
        min_value: float = data[0]
        if file_tag is None:
            file_tag = min_value
        else:
            file_tag = file_tag
        n: int = len(data)

        index = np.arange(1, n + 1) / n
        lorenz_curve = np.cumsum(data) / np.sum(data)
        plt.plot(index, lorenz_curve, color="orange", label="Lorenz Curve")
        plt.fill_between(index, lorenz_curve, index, color="orange", alpha=0.3)
        plt.xlabel(f"Cumulative Share of Population (truncated from {min_value})")
        plt.ylabel("Cumulative Share of Target Variable")
        gini: float = LorenzCurveGini.gini_coefficient(data)
        text_str: str = f"Gini: {gini:.4f}"
        plt.text(
            0.05,
            0.95,
            text_str,
            transform=plt.gca().transAxes,
            fontsize=10,
            verticalalignment="top",
        )
        if save_fig:
            plt.savefig(f"gini_from_{file_tag}.{file_type}", format=file_type)

    def lorenz_gini_from_to(
        self,
        lower_bound: Optional[float] = None,
        upper_bound: Optional[float] = None,
        save_fig: bool = False,
        file_type: str = "pdf",
    ) -> float:
        """
        Plot Lorenz curve from lower_bound to upper_bound.

        :param lower_bound: lower bound of the data
        :param upper_bound: upper bound of the data
        :param save_fig: whether to save the figure
        :return: Gini coefficient as a float.
        """
        data = self.slice_data(lower_bound, upper_bound)
        self.plot_lorenz_curve(
            data=data, save_fig=save_fig, file_tag=lower_bound, file_type=file_type
        )
        return self.gini_coefficient(data=data)

    def plot_lorenz_curves_with_lower_bounds(
        self,
        num_quantiles: int = 10,
        lower_bounds: Optional[List[float]] = None,
        slice_from: int = 0,
        duration: Optional[int] = 300,
    ) -> Tuple[List[float], List[Union[float, int]]]:
        """
        Plot Lorenz curves with lower bounds.

        :param num_quantiles: number of quantiles
        :param lower_bounds: lower bounds of the data
        :param slice_from: lower bound of the data
        :param duration: duration of the gif
        :return: a tuple of lists of floats and ints
        """
        if lower_bounds is None:
            bounds = self.get_bounds(
                num_quantiles, slice_from=slice_from, unique_bounds=True
            )
        elif lower_bounds is not None:
            bounds = lower_bounds
        gini_list: List[float] = []
        filenames: List[str] = []
        for bound in bounds:
            filename = f"gini_from_{bound}.png"
            filenames.append(filename)
            gini_list.append(
                self.lorenz_gini_from_to(
                    lower_bound=bound, save_fig=True, file_type="png"
                )
            )
            plt.close()

        with imageio.get_writer(
            f"lorenz_bounds_from_{bounds[0]}_to_{bounds[-1]}.gif",
            mode="I",
            duration=duration,
            loop=0,
        ) as writer:
            for filename in filenames:
                image = imageio.v2.imread(filename, pilmode="RGBA")
                writer.append_data(image)
        for filename in set(filenames):
            os.remove(filename)
        return (gini_list, bounds)

    def get_gini_list_from_bounds(self, bounds: List[float]) -> List[float]:
        """
        Calculate Gini coefficients from bounds.

        :param bounds: a list of lower bounds
        :return: a list of Gini coefficients
        """
        gini_list = []
        for bound in tqdm(bounds, desc="Processing bounds"):
            data = self.slice_data(bound, upper_bound=None)
            gini_list.append(self.gini_coefficient(data=data))
        return gini_list

    def cal_gini_from_quantile(self, quantile: float) -> float:
        """
        Calculate Gini coefficient from quantile.

        :param quantile: quantile of the data
        :return: Gini coefficient as a float.
        """
        lower_bound = np.quantile(self.data, quantile)
        data = self.slice_data(float(lower_bound), upper_bound=None)
        return self.gini_coefficient(data=data)

    def gini_lower_bounds_curve(
        self,
        num_quantiles: int = 100,
        lower_bounds: Optional[List[float]] = None,
        slice_from: int = 0,
    ) -> List[float]:
        """
        Calculate and plot Gini coefficients with lower_bounds list.

        :param num_quantiles: number of quantiles
        :param lower_bounds: a list of lower bounds
        :param slice_from: slice data from slice_from
        :return: a list of Gini coefficients
        """
        gini_list = []
        if lower_bounds is None:
            lower_bounds = self.get_bounds(num_quantiles, slice_from=slice_from)
        for lower_bound in lower_bounds:
            data = self.slice_data(lower_bound, upper_bound=None)
            gini_list.append(self.gini_coefficient(data=data))
        indice = np.arange(1, len(lower_bounds) + 1)

        _, ax1 = plt.subplots()
        ax1_color, ax2_color = "orange", "green"
        ax1.plot(indice, gini_list, color=ax1_color)
        ax2 = ax1.twinx()
        ax2.plot(indice, lower_bounds, color=ax2_color)
        ax1.tick_params(axis="y", colors=f"tab:{ax1_color}")
        ax2.tick_params(axis="y", colors=f"tab:{ax2_color}")
        ax1.set_xlabel("Quantiles")
        ax1.set_ylabel("Gini Coefficients", color=f"tab:{ax1_color}")
        ax2.set_ylabel("Lower Bounds", color=f"tab:{ax2_color}")
        plt.show()
        return gini_list



from collections import Counter
from typing import List, Optional, Union

import numpy as np
import pandas as pd

from .base import BaseSampler


class FrequencySampler(BaseSampler):
    """
    This class provides method to sample data by frequency.
    """

    def __init__(
        self,
        sample_size: int,
        data: Union[pd.Series, List[float]],
        slice_from: Optional[float] = None,
        slice_to: Optional[float] = None,
        log_scale: Optional[bool] = True,
        laplace_smoothing: Optional[bool] = True,
    ) -> None:
        super().__init__(
            sample_size, data, slice_from, slice_to, log_scale, laplace_smoothing
        )

    def sample(self) -> dict:
        if self.slice_from is not None:
            self.data = [x for x in self.data if x > self.slice_from]
        if self.slice_to is not None:
            self.data = [x for x in self.data if x <= self.slice_to]

        # calculate the percentiles that will give us the required sample size
        percentiles = np.linspace(0, 100, self.sample_size + 2)[1:-1]
        samples = np.percentile(self.data, percentiles)
        if self.log_scale:
            samples = [np.exp(x) for x in samples]
        else:
            samples = [x for x in samples]
        if self.laplace_smoothing:
            samples = [x - 1 for x in samples]
        return dict(Counter(samples))



from .lorenz_curve import LorenzCurveGini
from .plot_trisurf import TrisurfPloter
from .portfolio_curve import PortfolioPlotter
from .venn2 import Venn2Ploter
from .venn3 import Venn3Ploter
from .venn_base import BaseVennPloter

__all__ = [
    "LorenzCurveGini",
    "BaseVennPloter",
    "Venn3Ploter",
    "Venn2Ploter",
    "TrisurfPloter",
    "PortfolioPlotter",
]



from typing import List, Optional, Union

from ..evaluation import Calculator, LogarithmPCACalculator


def evaluate_targets(
    calculator: Union[Calculator, LogarithmPCACalculator],
    evaluator_flags: List[str],
    hyperparameters: List[Optional[float]],
    evaluator_propertys: List[Optional[str]],
    groupbys: List[Optional[str]],
    target_columns: List[str],
    weights: List[float],
) -> List[float]:
    targets = []
    for flag, hyperparameter, evaluator_property, groupby, target_column in zip(
        evaluator_flags, hyperparameters, evaluator_propertys, groupbys, target_columns
    ):
        if flag == "portfolio":
            _, concentration = calculator.calculate_portfolio_concentration(
                target_column=target_column,
                expected_return=hyperparameter,
            )
            targets.append(concentration)

        elif flag == "distinct_count_portfolio":
            (
                _,
                concentration,
            ) = calculator.calculate_distinct_count_portfolio_concentration(
                target_column=target_column,
                expected_coverage=hyperparameter,
            )
            targets.append(concentration)
        elif flag == "wuauc":
            wuauc = calculator.calculate_wuauc(
                groupby=groupby,
                label_column=target_column,
                weights_for_equation=weights,
            )
            targets.append(wuauc)
        elif flag == "auc":
            auc = calculator.calculate_wuauc(
                groupby=groupby,
                label_column=target_column,
                weights_for_equation=weights,
                auc=True,
            )
            targets.append(auc)
        elif flag == "woauc":
            woauc = calculator.calculate_woauc(
                groupby=groupby,
                target_column=target_column,
                weights_for_equation=weights,
            )
            targets.append(sum(woauc))
        elif flag == "logmse":
            mse = calculator.calculate_log_mse(
                target_column=target_column,
            )
            targets.append(mse)
        elif flag == "neg_rank_ratio":
            neg_rank_ratio = calculator.calculate_neg_rank_ratio(
                weights_for_equation=weights, label_column=target_column
            )
            targets.append(neg_rank_ratio)
        elif flag == "inverse_pairs":
            inverse_score = calculator.calculate_inverse_pair(
                calculator=calculator,
                weights_for_equation=weights,
                weights_type=evaluator_property,
            )
            targets.append(inverse_score)
        elif flag == "tau":
            tau = calculator.calculate_tau(
                groupby=groupby,
                target_column=target_column,
                num_bins=hyperparameter,
            )
            targets.append(tau)
    return targets



import logging
from typing import Optional, Union

import optuna
import pandas as pd
from mixician import SelfBalancingLogarithmPCACalculator

from ..evaluation import Calculator, LogarithmPCACalculator
from .base import BasePipeline

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(levelname)s - %(message)s",
    datefmt="%H:%M:%S",
)
logger = logging.getLogger(__name__)


class LogarithmPCAPipeline(BasePipeline):
    """Pipeline for processing and optimizing PCA with logarithmic transformations.

    This pipeline extends the `BasePipeline` class to implement a specific process for
    optimizing Principal Component Analysis (PCA) with logarithmic transformations,
    particularly focusing on self-balancing mechanisms.

    Attributes:
        file_type (str): Type of the file to load data from, supported types are 'csv' and 'xlsx'.
        dataframe (pd.DataFrame): The loaded dataset in a pandas DataFrame.
        calculator (LogarithmPCACalculator): Calculator for PCA operations.
        objective (MultipleObjective): The optimization objective.
    """

    def __init__(
        self,
        dataframe: Optional[pd.DataFrame] = None,
        config_path: Optional[str] = None,
        n_trials: int = 200,
    ) -> None:
        """Initializes the pipeline with configuration and trial settings.

        Args:
            config_path (Optional[str]): Path to the configuration file. Defaults to None.
            n_trials (int): Number of optimization trials to perform. Defaults to 200.
        """
        super().__init__(
            dataframe=dataframe, config_path=config_path, n_trials=n_trials
        )
        self.run()

    def _load_calculator(self) -> Union[Calculator, LogarithmPCACalculator]:
        """Initializes the PCA calculator with the loaded dataset."""
        pca_calculator = SelfBalancingLogarithmPCACalculator(
            dataframe=self.dataframe,
            config=self.config["Calculator"],
        )
        self.calculator = LogarithmPCACalculator(
            pca_calculator=pca_calculator,
        )
        return self.calculator

    def show_results(self) -> None:
        """Displays the results of the optimization process."""
        self.calculator.pca_calculator.update(
            pca_weights=self.objective.best_params,
        )
        logger.info("Best parameters for PCA with logarithmic transformations:")
        self.calculator.pca_calculator.get_weights()
        self.objective.build_logger()
        self.objective.logger.info(
            f"Best parameters: {self.calculator.pca_calculator.results}"
        )
        self.calculator.pca_calculator.show_equation()



from abc import ABC, abstractmethod
from typing import Optional, Tuple

import pandas as pd


class BaseVennPloter(ABC):
    """Base class for venn ploter."""

    def __init__(
        self, dataframe: pd.DataFrame, column1_name: str, column2_name: str
    ) -> None:
        self.df = dataframe
        self.column1_name = column1_name
        self.column2_name = column2_name
        self.column1_data = dataframe[column1_name]
        self.column2_data = dataframe[column2_name]
        self.df_len = len(dataframe)

    @abstractmethod
    def get_conditions(self) -> Tuple:
        """Returns the conditions for the venn diagram"""
        raise NotImplementedError("get_conditions() not implemented")

    @abstractmethod
    def get_group_count_ratios(self) -> Tuple:
        """Returns the ratios of the groups based on the count"""
        raise NotImplementedError("get_group_count_ratios() not implemented")

    @abstractmethod
    def get_group_value_ratios(
        self,
        value_column_name: str,
        lower_bound: Optional[float] = None,
        upper_bound: Optional[float] = None,
    ) -> Tuple:
        """Returns the ratios of the groups based on the value

        value_column_name: name of the column with the values
        lower_bound: lower bound of the values
        upper_bound: upper bound of the values
        """
        raise NotImplementedError("get_group_value_ratios() not implemented")

    @abstractmethod
    def plot_ratio_venn(
        self,
        ratios: tuple,
        save_fig: bool = False,
        file_tag: Optional[float] = None,
        file_type: str = "pdf",
    ) -> None:
        """Plots the venn diagram based on the ratios"""
        raise NotImplementedError("plot_ratio_venn() not implemented")

    def plot_count_ratio_venn(self) -> None:
        """
        Plots the venn diagram with the count ratios
        """
        ratios = self.get_group_count_ratios()
        self.plot_ratio_venn(ratios)

    def plot_value_ratio_venn(
        self,
        value_column_name: str,
        lower_bound: Optional[float] = None,
        upper_bound: Optional[float] = None,
        save_fig: bool = False,
        file_type: str = "pdf",
    ) -> None:
        """
        Plots the venn diagram with the value ratios

        value_column_name: name of the column with the values
        lower_bound: lower bound of the values
        upper_bound: upper bound of the values
        save_fig: whether to save the figure
        """

        ratios = self.get_group_value_ratios(
            value_column_name, lower_bound=lower_bound, upper_bound=upper_bound
        )
        self.plot_ratio_venn(
            ratios, save_fig=save_fig, file_tag=upper_bound, file_type=file_type
        )



import os
import platform
import subprocess


def get_logical_processors_count() -> int:
    """
    Fetch the count of logical processors based on the system type.

    Returns:
        int: Number of logical processors.

    Raises:
        ValueError: If the system is not Windows, Darwin (MacOS), or Linux.
    """

    system: str = platform.system()

    if system == "Windows":
        return int(os.environ["NUMBER_OF_PROCESSORS"])
    elif system == "Darwin":
        return int(
            subprocess.check_output(["sysctl", "-n", "hw.ncpu"]).decode("utf-8").strip()
        )
    elif system == "Linux":
        command: str = "grep -c ^processor /proc/cpuinfo"
        return int(subprocess.check_output(command, shell=True).decode("utf-8").strip())
    else:
        raise ValueError(f"Unsupported OS: {system}")



from typing import TYPE_CHECKING, List, Optional

import numpy as np
import pandas as pd
from sklearn.metrics import roc_auc_score

if TYPE_CHECKING:
    from .calculator import Calculator


def calculate_wuauc(
    calculator: "Calculator",
    groupby: Optional[str],
    weights_for_equation: List,
    weights_for_groups: Optional[pd.Series] = None,
    label_column: str = "label",
    auc: bool = False,
) -> float:
    """Calculate weighted user AUC.

    :param groupby: groupby column
    :param weights_for_equation: weights for equation
    :param weights_for_groups: weights for group
    :param label_column: label column
    :param auc: bool, optional, default: False
    :return: AUC/WUAUC/UAUC
    """
    if auc:
        result = float(
            roc_auc_score(
                calculator.df[label_column].values, calculator.df["overall_score"]
            )
        )
    else:
        if groupby is not None:
            grouped = calculator.df.groupby(groupby).apply(
                lambda x: float(roc_auc_score(x[label_column], x["overall_score"]))
            )
            if weights_for_groups is not None:
                counts_sorted = weights_for_groups.loc[grouped.index]
                result = float(np.average(grouped, weights=counts_sorted.values))
            else:
                result = float(np.mean(grouped))
    return result



from .base import BaseDataLoader
from .load_config import load_config
from .load_csv import CSVLoader
from .load_excel import ExcelLoader

__all__ = ["BaseDataLoader", "CSVLoader", "ExcelLoader", "load_config"]
